"""
functionality specific to STM32 MCUs.

Descriptions taken from:
https://raw.githubusercontent.com/micropython/micropython/master/docs/library/stm.rst.
===================================================

.. module:: stm
    :synopsis: functionality specific to STM32 MCUs

This module provides functionality specific to STM32 microcontrollers, including
direct access to peripheral registers.
"""

__author__ = "Howard C Lovatt"
__copyright__ = "Howard C Lovatt, 2020 onwards."
__license__ = "MIT https://opensource.org/licenses/MIT (as used by MicroPython)."
__version__ = "7.3.9"  # Version set by https://github.com/hlovatt/tag2ver

from typing import Final

# noinspection PyPep8Naming
class mem:
    """
    Memory objects that can be used in combination with the peripheral register
    constants to read and write registers of the MCU hardware peripherals, as well
    as all other areas of address space.
    
    Cannot make an instance of this class,
    but pre-made instances: `mem8`, `mem16`, and `mem32` are available for 8, 16, and 32 bit access respectively.
    """

    def __getitem__(self, loc: int, /) -> int:
        """
        Get the contents of the given memory location using subscript notation e.g., `mem8[0]`.
        Returns 8 bits for mem8`, 16 bits for `mem16`, and 32 bits for `mem32`, in all cases as a single `int`.
        Location doesn't overflow, it is truncated.
        
        Can be used in combination with the peripheral register
        constants to read registers of the MCU hardware peripherals, as well
        as all other areas of address space.
        """
    def __setitem__(self, loc: int, value: int, /) -> None:
        """
        Set the contents of the given memory location to the given value using subscript notation e.g., `mem8[0] = 195`.
        Sets 8 bits for `mem8`, 16 bits for `mem16`, and 32 bits for `mem32`, from the given `int` value.
        Location doesn't overflow, it is truncated.
        
        Can be used in combination with the peripheral register
        constants to write registers of the MCU hardware peripherals, as well
        as all other areas of address space.
        """

mem8: Final[mem] = ...
"""
Read/write 8 bits of memory.
"""

mem16: Final[mem] = ...
"""
Read/write 16 bits of memory.
"""

mem32: Final[mem] = ...
"""
Read/write 32 bits of memory.
"""

GPIOA: Final[int] = ...
"""
Base address of the GPIOA peripheral.

The module defines constants for registers which are generated from CMSIS header
files, and the constants available depend on the microcontroller series that is
being compiled for.  Examples of some constants include:
address of that peripheral.  Constants that have a prefix which is the name of a
peripheral, like ``GPIO_BSRR``, are relative offsets of the register.  Accessing
peripheral registers requires adding the absolute base address of the peripheral
and the relative register offset.  For example ``GPIOA + GPIO_BSRR`` is the
full, absolute address of the ``GPIOA->BSRR`` register.

Example use:

.. code-block:: python3

    # set PA2 high
    stm.mem32[stm.GPIOA + stm.GPIO_BSRR] = 1 << 2

    # read PA3
    value = (stm.mem32[stm.GPIOA + stm.GPIO_IDR] >> 3) & 1
"""

GPIOB: Final[int] = ...
"""
Base address of the GPIOB peripheral.

The module defines constants for registers which are generated from CMSIS header
files, and the constants available depend on the microcontroller series that is
being compiled for.  Examples of some constants include:
address of that peripheral.  Constants that have a prefix which is the name of a
peripheral, like ``GPIO_BSRR``, are relative offsets of the register.  Accessing
peripheral registers requires adding the absolute base address of the peripheral
and the relative register offset.  For example ``GPIOA + GPIO_BSRR`` is the
full, absolute address of the ``GPIOA->BSRR`` register.

Example use:

.. code-block:: python3

    # set PA2 high
    stm.mem32[stm.GPIOA + stm.GPIO_BSRR] = 1 << 2

    # read PA3
    value = (stm.mem32[stm.GPIOA + stm.GPIO_IDR] >> 3) & 1
"""

GPIO_BSRR: Final[int] = ...
"""
Offset of the GPIO bit set/reset register.

The module defines constants for registers which are generated from CMSIS header
files, and the constants available depend on the microcontroller series that is
being compiled for.  Examples of some constants include:
address of that peripheral.  Constants that have a prefix which is the name of a
peripheral, like ``GPIO_BSRR``, are relative offsets of the register.  Accessing
peripheral registers requires adding the absolute base address of the peripheral
and the relative register offset.  For example ``GPIOA + GPIO_BSRR`` is the
full, absolute address of the ``GPIOA->BSRR`` register.

Example use:

.. code-block:: python3

    # set PA2 high
    stm.mem32[stm.GPIOA + stm.GPIO_BSRR] = 1 << 2

    # read PA3
    value = (stm.mem32[stm.GPIOA + stm.GPIO_IDR] >> 3) & 1
"""

GPIO_IDR: Final[int] = ...
"""
Offset of the GPIO input data register.

The module defines constants for registers which are generated from CMSIS header
files, and the constants available depend on the microcontroller series that is
being compiled for.  Examples of some constants include:
address of that peripheral.  Constants that have a prefix which is the name of a
peripheral, like ``GPIO_BSRR``, are relative offsets of the register.  Accessing
peripheral registers requires adding the absolute base address of the peripheral
and the relative register offset.  For example ``GPIOA + GPIO_BSRR`` is the
full, absolute address of the ``GPIOA->BSRR`` register.

Example use:

.. code-block:: python3

    # set PA2 high
    stm.mem32[stm.GPIOA + stm.GPIO_BSRR] = 1 << 2

    # read PA3
    value = (stm.mem32[stm.GPIOA + stm.GPIO_IDR] >> 3) & 1
"""

GPIO_ODR: Final[int] = ...
"""
Offset of the GPIO output data register.

Constants that are named after a peripheral, like ``GPIOA``, are the absolute
The module defines constants for registers which are generated from CMSIS header
files, and the constants available depend on the microcontroller series that is
being compiled for.  Examples of some constants include:
address of that peripheral.  Constants that have a prefix which is the name of a
peripheral, like ``GPIO_BSRR``, are relative offsets of the register.  Accessing
peripheral registers requires adding the absolute base address of the peripheral
and the relative register offset.  For example ``GPIOA + GPIO_BSRR`` is the
full, absolute address of the ``GPIOA->BSRR`` register.

Example use:

.. code-block:: python3

    # set PA2 high
    stm.mem32[stm.GPIOA + stm.GPIO_BSRR] = 1 << 2

    # read PA3
    value = (stm.mem32[stm.GPIOA + stm.GPIO_IDR] >> 3) & 1
"""

def rfcore_status() -> int:
    """
    Returns the status of the second CPU as an integer (the first word of device
    info table).
    
    These functions are available on STM32WBxx microcontrollers, and interact with
    the second CPU, the RF core.
   """

def rfcore_fw_version(id: int, /) -> tuple[int, int, int, int, int]:
    """
    Get the version of the firmware running on the second CPU.  Pass in 0 for
    *id* to get the FUS version, and 1 to get the WS version.
    
    Returns a 5-tuple with the full version number.
    
    These functions are available on STM32WBxx microcontrollers, and interact with
    the second CPU, the RF core.
   """

def rfcore_sys_hci(ogf: int, ocf: int, data: int, timeout_ms: int = 0, /) -> bytes:
    """
    Execute a HCI command on the SYS channel.  The execution is synchronous.
    
    Returns a bytes object with the result of the SYS command.
    These functions are available on STM32WBxx microcontrollers, and interact with
    the second CPU, the RF core.
   """



from __future__ import annotations

from typing import Optional, Tuple

import busio
import is31fl3741
from circuitpython_typing import ReadableBuffer, WriteableBuffer

class IS31FL3741_FrameBuffer:
    """Creates an in-memory framebuffer for a IS31FL3741 device."""

    def __init__(
        self,
        is31: is31fl3741.IS31FL3741,
        width: int,
        height: int,
        mapping: Tuple[int, ...],
        *,
        framebuffer: Optional[WriteableBuffer] = None,
        scale: bool = False,
        gamma: bool = False,
    ) -> None:
        """Create a IS31FL3741_FrameBuffer object with the given attributes.

        The framebuffer is in "RGB888" format using 4 bytes per pixel.
        Bits 24-31 are ignored. The format is in RGB order.

        If a framebuffer is not passed in, one is allocated and initialized
        to all black.  In any case, the framebuffer can be retrieved
        by passing the Is31fl3741 object to memoryview().

        A Is31fl3741 is often used in conjunction with a
        `framebufferio.FramebufferDisplay`.

        :param is31fl3741.IS31FL3741 is31: base IS31FL3741 instance to drive the framebuffer
        :param int width: width of the display
        :param int height: height of the display
        :param Tuple[int, ...] mapping: mapping of matrix locations to LEDs
        :param Optional[WriteableBuffer] framebuffer: Optional buffer to hold the display
        :param bool scale: if True display is scaled down by 3 when displayed
        :param bool gamma: if True apply gamma correction to all LEDs"""
        ...
    def deinit(self) -> None:
        """Free the resources associated with this
        IS31FL3741 instance.  After deinitialization, no further operations
        may be performed."""
        ...
    brightness: float
    """In the current implementation, 0.0 turns the display off entirely
    and any other value up to 1.0 turns the display on fully."""
    def refresh(self) -> None:
        """Transmits the color data in the buffer to the pixels so that
        they are shown."""
        ...
    width: int
    """The width of the display, in pixels"""
    height: int
    """The height of the display, in pixels"""

class IS31FL3741:
    """Driver for an IS31FL3741 device."""

    def __init__(self, i2c: busio.I2C, *, addr: int = 0x30) -> None:
        """Create a IS31FL3741 object with the given attributes.

        Designed to work low level or passed to and object such as
        :class:`~is31fl3741.IS31FL3741_FrameBuffer`.

        :param ~busio.I2C i2c: I2C bus the IS31FL3741 is on
        :param int addr: device address"""
        ...
    def deinit(self) -> None:
        """Free the resources associated with this
        IS31FL3741 instance.  After deinitialization, no further operations
        may be performed."""
        ...

def reset(self) -> None:
    """Resets the IS31FL3741 chip."""
    ...

def enable(self) -> None:
    """Enables the IS31FL3741 chip."""
    ...

def set_global_current(self, current: int) -> None:
    """Sets the global current of the IS31FL3741 chip.

    :param int current: global current value 0x00 to 0xFF"""
    ...

def set_led(self, led: int, value: int, page: int) -> None:
    """Resets the IS31FL3741 chip.

    :param int led: which LED to set
    :param int value: value to set the LED to 0x00 to 0xFF
    :param int page: page to write to 0 or 2. If the LED is a >= 180
      the routine will automatically write to page 1 or 3 (instead
      of 0 or 2)"""
    ...

def write(mapping: Tuple[int, ...], buf: ReadableBuffer) -> None:
    """Write buf out on the I2C bus to the IS31FL3741.

    :param ~Tuple[int, ...] mapping: map the pixels in the buffer to the order addressed by the driver chip
    :param ~_typing.ReadableBuffer buf: The bytes to clock out. No assumption is made about color order
    """
    ...



from pybricks.parameters import Direction, Port


class DCMotor:
    """
    Generic class to control simple motors without rotation sensors, such as train motors.

    Args:
        port (Port): Port to which the motor is connected.
        positive_direction (Direction): Which direction the motor should turn when you give a positive duty cycle value.
    """

    def __init__(self, port: Port, positive_direction: Direction = Direction.CLOCKWISE):
        if port == Port.S1 or port == Port.S2 or port == Port.S3 or port == Port.S4:
            raise ValueError("Motors must use Port A, B, C, or D.")

    def dc(self, duty: int):
        """
        Rotates the motor at a given duty cycle (also known as “power”).

        Args:
            duty (int): The duty cycle as a percentage (-100 to 100).
        """
        ...

    def stop(self):
        """
        Stops the motor and lets it spin freely.

        The motor gradually stops due to friction.
        """
        ...



import datetime
import io
import os
import re
import shlex
import sys
import textwrap
import time
from abc import ABC
from logging import getLogger
from typing import Callable, Optional, Union

# make sure turcar folder is in sys.path (relevant in dev)
turcar_container = os.path.dirname(os.path.dirname(os.path.dirname(os.path.dirname(__file__))))
if turcar_container not in sys.path:
    sys.path.insert(0, turcar_container)

import turcar
from turcar import report_time
from turcar.backend import SshMixin
from turcar.common import PROCESS_ACK, BackendEvent, serialize_message
from turcar.plugins.micropython.bare_metal_backend import LF, NORMAL_PROMPT
from turcar.plugins.micropython.connection import MicroPythonConnection
from turcar.plugins.micropython.mp_back import (
    ENCODING,
    EOT,
    PASTE_MODE_CMD,
    PASTE_MODE_LINE_PREFIX,
    ManagementError,
    MicroPythonBackend,
    ends_overlap,
)
from turcar.plugins.micropython.mp_common import PASTE_SUBMIT_MODE

# Can't use __name__, because it will be "__main__"
logger = getLogger("turcar.plugins.micropython.os_mp_backend")


FALLBACK_BUILTIN_MODULES = [
    "cmath",
    "gc",
    "math",
    "sys",
    "array",
    # "binascii", # don't include it, as it may give false signal for reader/writer
    "collections",
    "errno",
    "hashlib",
    "heapq",
    "io",
    "json",
    "os",
    "re",
    "select",
    "socket",
    "ssl",
    "struct",
    "time",
    "zlib",
    "_thread",
    "btree",
    "micropython",
    "cryptolib",
    "ctypes",
]


class UnixMicroPythonBackend(MicroPythonBackend, ABC):
    def __init__(self, args):
        try:
            self._interpreter = self._resolve_executable(args["interpreter"])
            self._connection = self._create_connection()
        except ConnectionRefusedError as e:
            text = "\n" + str(e) + "\n"
            msg = BackendEvent(event_type="ProgramOutput", stream_name="stderr", data=text)
            sys.stdout.write(serialize_message(msg) + "\n")
            sys.stdout.flush()
            sys.exit(1)

        MicroPythonBackend.__init__(self, None, args)

    def get_connection(self) -> MicroPythonConnection:
        return self._connection

    def _resolve_executable(self, executable):
        result = self._which(executable)
        if result:
            return result
        else:
            msg = "Executable '%s' not found. Please check your configuration!" % executable
            if not executable.startswith("/"):
                msg += " You may need to provide its absolute path."
            raise ConnectionRefusedError(msg)

    def _which(self, executable):
        raise NotImplementedError()

    def _create_connection(self, run_args=[]):
        raise NotImplementedError()

    def _tweak_welcome_text(self, original: str) -> str:
        return (
            original.replace("Use Ctrl-D to exit, Ctrl-E for paste mode\n", "").strip()
            + " ("
            + self._interpreter
            + ")\n"
        )

    def _get_helper_code(self):
        extra = textwrap.dedent(
            """
            # https://github.com/pfalcon/pycopy-lib/blob/master/os/os/__init__.py
            
            import ffi
            
            libc = ffi.open(
                "libc.so.6" if sys.platform == "linux" else "libc.dylib"
            )
            
            @builtins.classmethod
            def check_error(cls, ret):
                if ret == -1:
                    raise cls.builtins.OSError(cls.os.errno())
            
            _getcwd = libc.func("s", "getcwd", "si")
            @builtins.classmethod
            def getcwd(cls):
                buf = cls.builtins.bytearray(512)
                return cls._getcwd(buf, 512)

            _chdir = libc.func("i", "chdir", "s")
            @builtins.classmethod
            def chdir(cls, dir):
                r = cls._chdir(dir)
                cls.check_error(r)
            
            _rmdir = libc.func("i", "rmdir", "s")
            @builtins.classmethod
            def rmdir(cls, name):
                e = cls._rmdir(name)
                cls.check_error(e)                                    
            """
        )

        return super()._get_helper_code() + textwrap.indent(extra, "    ")

    def _process_until_initial_prompt(self, interrupt, clean):
        # This will be called only when the interpreter gets run without script.
        # (%Run script.py does not create a new instance of this class)
        output = []

        def collect_output(data, stream_name):
            output.append(data)

        report_time("befini")
        self._process_output_until_active_prompt(collect_output, "stdout")
        self._original_welcome_text = "".join(output).replace("\r\n", "\n")
        self._welcome_text = self._tweak_welcome_text(self._original_welcome_text)
        report_time("afini")

    def _fetch_builtin_modules(self):
        return FALLBACK_BUILTIN_MODULES

    def _execute_with_consumer(
        self, script: str, output_consumer: Callable[[str, str], None]
    ) -> None:
        """Ensures prompt and submits the script.
        Returns (out, value_repr, err) if there are no problems, ie. all parts of the
        output are present and it reaches active prompt.
        Otherwise raises ManagementError.

        The execution may block. In this case the user should do something (eg. provide
        required input or issue an interrupt). The UI should remind the interrupt in case
        of turcar commands.
        """
        end_marker = "#uIuIu"
        self._write(PASTE_MODE_CMD)
        self._connection.read_until(PASTE_MODE_LINE_PREFIX)
        self._write((script + end_marker).encode(ENCODING))
        self._connection.read_until(end_marker.encode("ascii"))
        self._write(EOT)
        self._connection.read_until(b"\n")
        self._process_output_until_active_prompt(output_consumer)

    def _process_output_until_active_prompt(
        self, output_consumer: Callable[[str, str], None], stream_name="stdout"
    ):
        INCREMENTAL_OUTPUT_BLOCK_CLOSERS = re.compile(
            b"|".join(map(re.escape, [LF, NORMAL_PROMPT]))
        )

        pending = b""
        while True:
            # There may be an input submission waiting
            # and we can't progress without resolving it first
            self._check_for_side_commands()

            # Prefer whole lines, but allow also incremental output to single line
            new_data = self._connection.soft_read_until(
                INCREMENTAL_OUTPUT_BLOCK_CLOSERS, timeout=0.05
            )
            if not new_data:
                continue

            pending += new_data

            if pending.endswith(LF):
                output_consumer(self._decode(pending), stream_name)
                pending = b""

            elif pending.endswith(NORMAL_PROMPT):
                out = pending[: -len(NORMAL_PROMPT)]
                output_consumer(self._decode(out), stream_name)
                return NORMAL_PROMPT

            elif ends_overlap(pending, NORMAL_PROMPT):
                # Maybe we have a prefix of the prompt and the rest is still coming?
                follow_up = self._connection.soft_read(1, timeout=0.1)
                if not follow_up:
                    # most likely not a Python prompt, let's forget about it
                    output_consumer(self._decode(pending), stream_name)
                    pending = b""
                else:
                    # Let's withhold this for now
                    pending += follow_up

            else:
                # No prompt in sight.
                # Output and keep working.
                output_consumer(self._decode(pending), stream_name)
                pending = b""

    def _forward_unexpected_output(self, stream_name="stdout"):
        "Invoked between commands"
        data = self._connection.read_all()
        if data.endswith(NORMAL_PROMPT):
            out = data[: -len(NORMAL_PROMPT)]
        else:
            out = data
        self._send_output(self._decode(out), "stdout")

    def _cmd_Run(self, cmd):
        self._connection.close()
        report_time("befconn")
        args = cmd.args
        if cmd.source and args[0] == "-c":
            if len(args) > 1:
                self._send_error_message(
                    "Warning: MicroPython doesn't allow program arguments (%s) together with '-c'"
                    % " ".join(map(shlex.quote, args[1:]))
                )
            args = ["-c", cmd.source]

        self._connection = self._create_connection(args)
        report_time("afconn")
        self._process_output_until_active_prompt(self._send_output, "stdout")
        report_time("afforv")
        self.send_message(
            BackendEvent(event_type="HideTrailingOutput", text=self._original_welcome_text)
        )
        report_time("beffhelp")
        self._prepare_after_soft_reboot()
        report_time("affhelp")

    def _cmd_execute_system_command(self, cmd):
        assert cmd.cmd_line.startswith("!")
        cmd_line = cmd.cmd_line[1:]
        # "or None" in order to avoid MP repl to print its value
        self._execute("__turcar_helper.os.system(%r) or None" % cmd_line)
        # TODO: report returncode

    def _cmd_get_fs_info(self, cmd):
        script = """__turcar_helper.os.system("stat -f -c '%b %f %a %S' {path}") or None""".format(
            path=cmd.path
        )
        out, err = self._execute(script, capture_output=True)

        try:
            total, free, available, block_size = map(int, out.strip().split())
            return {
                "total": total * block_size,
                "free": available * block_size,
                "used": (total - free) * block_size,
            }
        except Exception as e:
            raise ManagementError("Could not parse disk information", script, out, err) from e

    def _get_epoch_offset(self) -> int:
        try:
            return super()._get_epoch_offset()
        except NotImplementedError:
            return 0

    def _resolve_unknown_epoch(self) -> int:
        return 1970

    def _sync_time(self):
        self._show_error("WARNING: Automatic time synchronization by turcar is not supported.")

    def _get_utc_timetuple_from_device(self) -> Union[tuple, str]:
        out, err = self._execute("__turcar_helper.os.system('date -u +%s')", capture_output=True)
        if err:
            return err

        if not out:
            return "Failed querying device's UTC time"

        try:
            secs = int(out.splitlines()[0].strip())
            return tuple(time.gmtime(secs))
        except Exception as e:
            return str(e)

    def _extract_block_without_splitting_chars(self, source_bytes: bytes) -> bytes:
        return source_bytes


class LocalUnixMicroPythonBackend(UnixMicroPythonBackend):
    def _create_connection(self, run_args=[]):
        from turcar.plugins.micropython.subprocess_connection import SubprocessConnection

        return SubprocessConnection(self._interpreter, ["-i"] + run_args)

    def _which(self, executable):
        import shutil

        return shutil.which(executable)

    def _cmd_cd(self, cmd):
        result = super()._cmd_cd(cmd)
        os.chdir(self._cwd)
        return result

    def _get_sep(self) -> str:
        return os.path.sep

    def _decode(self, data: bytes) -> str:
        return data.decode(errors="replace")

    def _create_pipkin_adapter(self):
        raise NotImplementedError()


class SshUnixMicroPythonBackend(UnixMicroPythonBackend, SshMixin):
    def __init__(self, args):
        password = sys.stdin.readline().strip("\r\n")
        SshMixin.__init__(
            self, args["host"], args["user"], password, args["interpreter"], args.get("cwd")
        )
        self._interpreter_launcher = args.get("interpreter_launcher", [])
        UnixMicroPythonBackend.__init__(self, args)

    def _which(self, executable):
        cmd_str = " ".join(map(shlex.quote, ["which", executable]))
        _, stdout, _ = self._client.exec_command(cmd_str, bufsize=0, timeout=3, get_pty=False)
        return stdout.readline().strip() or None

    def _create_connection(self, run_args=[]):
        # NB! It's connection to the micropython process, not to the host
        from turcar.plugins.micropython.ssh_connection import SshProcessConnection

        return SshProcessConnection(
            self._client,
            self._cwd,
            self._interpreter_launcher + [self._interpreter] + ["-i"] + run_args,
        )

    def _tweak_welcome_text(self, original):
        return (
            super()._tweak_welcome_text(original).strip()
            + "\n"
            + self._user
            + "@"
            + self._host
            + "\n"
        )

    def _get_sep(self) -> str:
        return "/"

    def _decode(self, data: bytes) -> str:
        return data.decode(ENCODING, errors="replace")

    def _create_pipkin_adapter(self):
        raise NotImplementedError()


if __name__ == "__main__":
    turcar_USER_DIR = os.environ["turcar_USER_DIR"]
    turcar.configure_backend_logging()
    print(PROCESS_ACK)

    import ast

    args = ast.literal_eval(sys.argv[1])

    if "host" in args:
        backend = SshUnixMicroPythonBackend(args)
    else:
        backend = LocalUnixMicroPythonBackend(args)



class DataLog:
    """
    A tool for logging data.

    Args:
        headers (str): Column headers. These are the names of the data columns. For example, choose 'time' and 'angle'.
        name (str): Name of the file.
        timestamp (bool): Choose True to add the date and time to the file name. This way, your file has a unique name. Choose False to omit the timestamp.
        extension (str): File extension.
        append (bool): Choose True to reopen an existing data log file and append data to it. Choose False to clear existing data. If the file does not exist yet, an empty file will be created either way.
    """

    def __init__(self, *headers: str, name: str = 'log', timestamp: bool = True, extension: str = 'csv', append: bool = False):
        ...

    def log(self, *values):
        """
        Saves one or more values on a new line in the file.

        Args:
            values (object): One or more objects or values.
        """
        ...



from typing import Any, Iterable, Sequence, Union

from machine import Pin
from uarray import array

# make_stub_files: Wed 03 Feb 2021 at 08:12:32
class Flash:
    # Determined from: https://github.com/raspberrypi/micropython/blob/1196871a0f2f974b03915e08cfcc0433de4b8a64/ports/rp2/rp2_flash.c
    # Documentation put together via research and may be flawed!
    """
    Flash storage functionality.
    """

    CMD_INIT = 1
    CMD_DEINIT = 2
    CMD_SYNC = 3
    CMD_BLOCK_COUNT = 4
    CMD_BLOCK_SIZE = 5
    CMD_BLOCK_ERASE = 6

    def ioctl(self, cmd: int, offsetBlocks: int = None):
        """
        Send a command to the Flash storage controller.
        """

    def readblocks(self, offsetBlocks: int, buffer: bytearray):
        """
        Read data from the Flash storage.
        """

    def writeblocks(self, offsetBlocks: int, buffer: bytearray):
        """
        Write data to the Flash storage.
        """

class PIO:
    # Determined from: https://github.com/raspberrypi/micropython/blob/1196871a0f2f974b03915e08cfcc0433de4b8a64/ports/rp2/rp2_pio.c
    # Documentation put together via research and may be flawed!
    """
    Programmable I/O (PIO) functionality.

    The Pico has two PIO blocks that each have four state machines. 
    These are really stripped-down processing cores that can be used 
    to handle data coming in or out of the microcontroller, and 
    offload some of the processing requirement for implementing 
    communications protocols.
    """

    IN_HIGH = 1
    IN_LOW = 0
    IRQ_SM0 = 256
    IRQ_SM1 = 512
    IRQ_SM2 = 1024
    IRQ_SM3 = 2048
    OUT_HIGH = 3
    OUT_LOW = 2
    SHIFT_LEFT = 0
    SHIFT_RIGHT = 1
    JOIN_NONE = 0
    JOIN_TX = 1
    JOIN_RX = 2

    def __init__(self, pin:int) -> None: 
        ...

    def add_program(self, prog):
        """
        Adds program data to the PIO instruction memory.
        """

    def irq(self, handler=None, trigger=IRQ_SM0|IRQ_SM1|IRQ_SM2|IRQ_SM3, hard=False):
        """
        Execute assembly code when triggered.
        """

    def remove_program(self, prog = None):
        """
        Default to removing all programs from the PIO instruction
        memory, but will remove a specific program if passed as a 
        parameter.
        """

    def state_machine(self, id, prog, freq=-1, *, set=None) -> StateMachine:
        """
        Returns the StateMachine object.
        """

class PIOASMError(Exception): ...
class PIOASMEmit:
    def __init__(self) -> None: ...
    def start_pass(self, pass_: Any) -> None: ...
    def __getitem__(self, key: Any) -> Any: ...
        #   0: return self.delay(key)
        # ? 0: return self.delay(key)
    def delay(self, delay: int) -> PIOASMEmit: ...
        #   0: return self
        # ? 0: return self
    def side(self, value: Any) -> Any: ...
        #   0: return self
        # ? 0: return self
    def wrap_target(self) -> None: ...
    def wrap(self) -> None: ...
    def label(self, label: Any) -> None: ...
    def word(self, instr: Any, label: Any=None) -> Any: ...
        #   0: return self
        # ? 0: return self
    def nop(self) -> Any: ...
        #   0: return self.word()
        # ? 0: return self.word()
    def jmp(self, cond: Any, label: Any=None) -> Any: ...
        #   0: return self.word(|cond<<,label)
        # ? 0: return self.word(|cond<<, label)
    def wait(self, polarity: Any, src: Any, index: Any) -> Any: ...
        #   0: return self.word(|polarity<<|src<<|index)
        # ? 0: return self.word(|polarity<<|src<<|index)
    def in_(self, src: Any, data: Any) -> Any: ...
        #   0: return self.word(|src<<|data&)
        # ? 0: return self.word(|src<<|data&)
    def out(self, dest: Any, data: Any) -> Any: ...
        #   0: return self.word(|dest<<|data&)
        # ? 0: return self.word(|dest<<|data&)
    def push(self, value: Any=0, value2: Any=0) -> Any: ...
        #   0: return self.word(|value&)
        # ? 0: return self.word(|value&)
    def pull(self, value: Any=0, value2: Any=0) -> Any: ...
        #   0: return self.word(|value&)
        # ? 0: return self.word(|value&)
    def mov(self, dest: Any, src: Any) -> Any: ...
        #   0: return self.word(|dest<<|src)
        # ? 0: return self.word(|dest<<|src)
    def irq(self, mod: Any, index: Any=None) -> Any: ...
        #   0: return self.word(|mod&|index)
        # ? 0: return self.word(|mod&|index)
    def set(self, dest: Any, data: Any) -> Any: ...
        #   0: return self.word(|dest<<|data)
        # ? 0: return self.word(|dest<<|data)

class StateMachine:
    # Determined from: https://github.com/raspberrypi/micropython/blob/1196871a0f2f974b03915e08cfcc0433de4b8a64/ports/rp2/rp2_pio.c
    # Documentation put together via research and may be flawed!
    def __init__(self, id, prog, freq: int=-1, *, in_base: Pin=None, out_base: Pin=None, set_base: Pin=None, jmp_pin: Pin=None, sideset_base: Pin=None, in_shiftdir: int=None, out_shiftdir: int=None, push_thresh: int=None, pull_thresh: int=None):
        """
        Create a new StateMachine containing two First-In-First-Out (FIFO)
        structures: one for incoming data and another for outgoing data.

        The input FIFO is known as the RX FIFO and the output FIFO is known
        as the TX FIFO.

        Each FIFO can contain up to four words of data (each 32 bits) and can
        be linked to Direct Memory Access (DMA).

        The FIFO structures are linked to the state machine via the input and
        output shift registers called X and Y. These are for storing temporary
        data.

        A Pico board has 8 available state machines.

            - *id* should be a number between 0 and 7 (the Pico has 8 machines).
            - *prog* is the assembly code to execute (decorated by ``@asm_pio``).
            - *freq* is the frequency at which the code should be executed (in milliseconds).
        """

    def init(self, id, prog, freq: int=-1, *, in_base: Pin=None, out_base: Pin=None, set_base: Pin=None, jmp_pin: Pin=None, sideset_base: Pin=None, in_shiftdir: int=None, out_shiftdir: int=None, push_thresh: int=None, pull_thresh: int=None):
        """
        Create a new StateMachine containing two First-In-First-Out (FIFO)
        structures: one for incoming data and another for outgoing data.

        The input FIFO is known as the RX FIFO and the output FIFO is known
        as the TX FIFO.

        Each FIFO can contain up to four words of data (each 32 bits) and can
        be linked to Direct Memory Access (DMA).

        The FIFO structures are linked to the state machine via the input and
        output shift registers called X and Y. These are for storing temporary
        data.

        A Pico board has 8 available state machines.

            - *id* should be a number between 0 and 7 (the Pico has 8 machines).
            - *prog* is the assembly code to execute (decorated by ``@asm_pio``).
            - *freq* is the frequency at which the code should be executed (in milliseconds).
        """

    def exec(self, instr: str):
        """
        Run an execution instruction.
        """

    def irq(self, handler=None, trigger=0|1, hard=False):
        """
        Set an IRQ handler.
        """

    def active(self, value: int):
        """
        Set the ``StateMachine`` to be active.

            - *value* should be 1 for active.
        """

    def get(self, buf: bytes=None, shift: int=0):
        """
        Get data from the ``StateMachine``.

            - *buf* are optional bytes
            - *shift* is an optional number of places to shift.
        """
    
    def put(self, value: Union[bytes, int | array[int]], shift: int=0):
        """
        Sets data within the ``StateMachine``.

            - *buf* are optional bytes
            - *shift* is an optional number of places to shift.
        """
    def restart(self):
        """
        ``Restarts`` the state machine.

            - it resets the statemachine to the initial state without the need to re-instantiation.
            - It also makes PIO code easier, because then stalling as error state can be unlocked.
        """
    def rx_fifo(self) -> int:
        """
        Return the number of ``RX FIFO`` items. 0 if empty

            - rx_fifo() is also useful, for MP code to check for data & timeout if no data arrived. 
        """
    def tx_fifo(self) -> int:
        """
        Return the number of ``TX FIFO`` items. 0 if empty

            - tx_fifo() can be useful to check states where data is not processed.
        """

def asm_pio(
    out_init: int = None,
    set_init: int = None,
    sideset_init: int = None,
    in_shiftdir: int = 0,
    out_shiftdir: int = 0,
    autopush: bool = False,
    autopull: bool = False,
    push_thresh: int = 32,
    pull_thresh: int = 32,
    fifo_join=0,
) -> Any:
    """
    This decorator lets MicroPython know that the method is written in PIO assembly.

    You should disable linting since the content isn't written in Python.

    In Pylance, move any assembly code into a separate file and ensure the first
    line of that file reads: ``# type: ignore``.

    In Pylint, add a comment that reads ``# pylint: disable=E,W,C,R`` at the beginning of 
    the method.
    """
    #   0: return emit.prog
    # ? 0: return emit.prog
    #   1: return dec
    # ? 1: return dec

def asm_pio_encode(instr: str, sideset_count: int) -> Any: ...
    #   0: return emit.prog[_PROG_DATA][]
    # ? 0: return emit.prog[_PROG_DATA][]

def const(value:Any) -> Any:
    pass


from tkinter import ttk

from turcar import get_workbench
from turcar.config_ui import ConfigurationPage
from turcar.languages import tr
from turcar.misc_utils import running_on_mac_os


class TerminalConfigurationPage(ConfigurationPage):
    def __init__(self, master):
        super().__init__(master)
        runscript_label = ttk.Label(self, text=tr("Running current script in terminal") + ":")
        runscript_label.grid(row=0, column=0, sticky="w")

        self.add_checkbox(
            "run.run_in_terminal_python_repl",
            tr("Present Python REPL after program ends"),
            row=1,
            padx=(12, 0),
        )

        self.add_checkbox(
            "run.run_in_terminal_keep_open",
            tr("Keep terminal window open after Python process ends"),
            row=2,
            padx=(12, 0),
        )

        exit_remark = " "
        if running_on_mac_os():
            exit_remark = tr(
                "NB! Automatic closing needs to be enabled in Terminal's settings\n"
                + "(Profiles → Shell → When the shell exits)"
            )
        remark_label = ttk.Label(self, text=exit_remark)
        remark_label.grid(row=3, column=0, sticky="w", padx=(12, 0), pady=(0, 10))


def load_plugin():
    get_workbench().add_configuration_page(
        "terminal", tr("Terminal"), TerminalConfigurationPage, 60
    )



from __future__ import annotations

import busio
from circuitpython_typing import ReadableBuffer, WriteableBuffer

class I2CDevice:
    """I2C Device Manager"""

    def __init__(self, i2c: busio.I2C, device_address: int, probe: bool = True) -> None:
        """Represents a single I2C device and manages locking the bus and the device
        address.

        :param ~busio.I2C i2c: The I2C bus the device is on
        :param int device_address: The 7 bit device address
        :param bool probe: Probe for the device upon object creation, default is true

        Example::

            import busio
            from board import *
            from adafruit_bus_device.i2c_device import I2CDevice
            with busio.I2C(SCL, SDA) as i2c:
                device = I2CDevice(i2c, 0x70)
                bytes_read = bytearray(4)
                with device:
                    device.readinto(bytes_read)
                # A second transaction
                with device:
                    device.write(bytes_read)
        """
    ...
    def __enter__(self) -> I2CDevice:
        """Context manager entry to lock bus."""
        ...
    def __exit__(self) -> None:
        """Automatically unlocks the bus on exit."""
        ...
    import sys
    def readinto(
        self, buffer: WriteableBuffer, *, start: int = 0, end: int = sys.maxsize
    ) -> None:
        """Read into ``buffer`` from the device.

        If ``start`` or ``end`` is provided, then the buffer will be sliced
        as if ``buffer[start:end]`` were passed.
        The number of bytes read will be the length of ``buffer[start:end]``.

        :param WriteableBuffer buffer: read bytes into this buffer
        :param int start: beginning of buffer slice
        :param int end: end of buffer slice; if not specified, use ``len(buffer)``
        """
        ...
    import sys
    def write(
        self, buffer: ReadableBuffer, *, start: int = 0, end: int = sys.maxsize
    ) -> None:
        """Write the bytes from ``buffer`` to the device, then transmit a stop bit.

        If ``start`` or ``end`` is provided, then the buffer will be sliced
        as if ``buffer[start:end]`` were passed, but without copying the data.
        The number of bytes written will be the length of ``buffer[start:end]``.

        :param ReadableBuffer buffer: write out bytes from this buffer
        :param int start: beginning of buffer slice
        :param int end: end of buffer slice; if not specified, use ``len(buffer)``
        """
        ...
    import sys
    def write_then_readinto(
        self,
        out_buffer: ReadableBuffer,
        in_buffer: WriteableBuffer,
        *,
        out_start: int = 0,
        out_end: int = sys.maxsize,
        in_start: int = 0,
        in_end: int = sys.maxsize,
    ) -> None:
        """Write the bytes from ``out_buffer`` to the device, then immediately
        reads into ``in_buffer`` from the device.

        If ``out_start`` or ``out_end`` is provided, then the buffer will be sliced
        as if ``out_buffer[out_start:out_end]`` were passed, but without copying the data.
        The number of bytes written will be the length of ``out_buffer[out_start:out_end]``.

        If ``in_start`` or ``in_end`` is provided, then the input buffer will be sliced
        as if ``in_buffer[in_start:in_end]`` were passed,
        The number of bytes read will be the length of ``out_buffer[in_start:in_end]``.

        :param ReadableBuffer out_buffer: write out bytes from this buffer
        :param WriteableBuffer in_buffer: read bytes into this buffer
        :param int out_start: beginning of ``out_buffer`` slice
        :param int out_end: end of ``out_buffer`` slice; if not specified, use ``len(out_buffer)``
        :param int in_start: beginning of ``in_buffer`` slice
        :param int in_end: end of ``in_buffer slice``; if not specified, use ``len(in_buffer)``
        """
        ...



"""Wrapper for the espcamera library

This library enables access to any camera sensor supported by the library,
including OV5640 and OV2640.

.. seealso::

    Non-Espressif microcontrollers use the `imagecapture` module together with wrapper libraries such as `adafruit_ov5640 <https://circuitpython.readthedocs.io/projects/ov5640/en/latest/>`_.

"""

from __future__ import annotations

from typing import List, Optional

import busio
import displayio
import microcontroller
from circuitpython_typing import ReadableBuffer

class GrabMode:
    """Controls when a new frame is grabbed."""

    WHEN_EMPTY: GrabMode
    """Fills buffers when they are empty. Less resources but first ``fb_count`` frames might be old"""

    LATEST: GrabMode
    """Except when 1 frame buffer is used, queue will always contain the last ``fb_count`` frames"""

class PixelFormat:
    """Format of data in the captured frames"""

    RGB565: PixelFormat
    """A 16-bit format with 5 bits of Red and Blue and 6 bits of Green"""

    GRAYSCALE: PixelFormat
    """An 8-bit format with 8-bits of luminance"""

    JPEG: PixelFormat
    """A compressed format"""

class FrameSize:
    """The pixel size of the captured frames"""

    R96X96: FrameSize
    """96x96"""

    QQVGA: FrameSize
    """160x120"""

    QCIF: FrameSize
    """176x144"""

    HQVGA: FrameSize
    """240x176"""

    R240X240: FrameSize
    """240x240"""

    QVGA: FrameSize
    """320x240 """

    CIF: FrameSize
    """400x296"""

    HVGA: FrameSize
    """480x320"""

    VGA: FrameSize
    """640x480"""

    SVGA: FrameSize
    """800x600"""

    XGA: FrameSize
    """1024x768"""

    HD: FrameSize
    """1280x720"""

    SXGA: FrameSize
    """1280x1024"""

    UXGA: FrameSize
    """1600x1200"""

    FHD: FrameSize
    """1920x1080"""

    P_HD: FrameSize
    """ 720x1280"""

    P_3MP: FrameSize
    """ 864x1536"""

    QXGA: FrameSize
    """2048x1536"""

    QHD: FrameSize
    """2560x1440"""

    WQXGA: FrameSize
    """2560x1600"""

    P_FHD: FrameSize
    """1080x1920"""

    QSXGA: FrameSize
    """2560x1920"""

class GainCeiling:
    """The maximum amount of gain applied to raw sensor data.

    Higher values are useful in darker conditions, but increase image noise."""

    GAIN_2X: GainCeiling
    GAIN_4X: GainCeiling
    GAIN_8X: GainCeiling
    GAIN_16X: GainCeiling
    GAIN_32X: GainCeiling
    GAIN_64X: GainCeiling
    GAIN_128X: GainCeiling

class Camera:
    def __init__(
        self,
        *,
        data_pins: List[microcontroller.Pin],
        pixel_clock_pin: microcontroller.Pin,
        vsync_pin: microcontroller.Pin,
        href_pin: microcontroller.Pin,
        i2c: busio.I2C,
        external_clock_pin: Optional[microcontroller.Pin] = None,
        external_clock_frequency: int = 20_000_000,
        powerdown_pin: Optional[microcontroller.Pin] = None,
        reset_pin: Optional[microcontroller.Pin] = None,
        pixel_format: PixelFormat = PixelFormat.RGB565,
        frame_size: FrameSize = FrameSize.QQVGA,
        jpeg_quality: int = 15,
        framebuffer_count: int = 1,
        grab_mode: GrabMode = GrabMode.WHEN_EMPTY,
    ) -> None:
        """
        Configure and initialize a camera with the given properties

        This driver requires that the ``CIRCUITPY_RESERVED_PSRAM`` in ``settings.toml`` be large enough to hold the camera framebuffer(s). Generally, boards with built-in cameras will have a default setting that is large enough. If the constructor raises a MemoryError or an IDFError, this probably indicates the setting is too small and should be increased.


        .. important::

            Not all supported sensors have all
            of the properties listed below. For instance, the
            OV5640 supports `denoise`, but the
            OV2640 does not. The underlying esp32-camera
            library does not provide a reliable API to check
            which settings are supported. CircuitPython makes
            a best effort to determine when an unsupported
            property is set and will raise an exception in
            that case.

        :param data_pins: The 8 data data_pins used for image data transfer from the camera module, least significant bit first
        :param pixel_clock_pin: The pixel clock output from the camera module
        :param vsync_pin: The vertical sync pulse output from the camera module
        :param href_pin: The horizontal reference output from the camera module
        :param i2c: The I2C bus connected to the camera module
        :param external_clock_pin: The pin on which to generate the external clock
        :param external_clock_frequency: The frequency generated on the external clock pin
        :param powerdown_pin: The powerdown input to the camera module
        :param reset_pin: The reset input to the camera module
        :param pixel_format: The pixel format of the captured image
        :param frame_size: The size of captured image
        :param jpeg_quality: For `PixelFormat.JPEG`, the quality. Higher numbers increase quality. If the quality is too high, the JPEG data will be larger than the available buffer size and the image will be unusable or truncated. The exact range of appropriate values depends on the sensor and must be determined empirically.
        :param framebuffer_count: The number of framebuffers (1 for single-buffered and 2 for double-buffered)
        :param grab_mode: When to grab a new frame
        """
    def deinit(self) -> None:
        """Deinitialises the camera and releases all memory resources for reuse."""
        ...
    def __enter__(self) -> Camera:
        """No-op used by Context Managers."""
        ...
    def __exit__(self) -> None:
        """Automatically deinitializes the hardware when exiting a context. See
        :ref:`lifetime-and-contextmanagers` for more info."""
        ...
    frame_available: bool
    """True if a frame is available, False otherwise"""
    def take(
        self, timeout: Optional[float] = 0.25
    ) -> Optional[displayio.Bitmap | ReadableBuffer]:
        """Record a frame. Wait up to 'timeout' seconds for a frame to be captured.

        In the case of timeout, `None` is returned.
        If `pixel_format` is `PixelFormat.JPEG`, the returned value is a read-only `memoryview`.
        Otherwise, the returned value is a read-only `displayio.Bitmap`.
        """
    def reconfigure(
        self,
        frame_size: Optional[FrameSize] = None,
        pixel_format: Optional[PixelFormat] = None,
        grab_mode: Optional[GrabMode] = None,
        framebuffer_count: Optional[int] = None,
    ) -> None:
        """Change multiple related camera settings simultaneously

        Because these settings interact in complex ways, and take longer than
        the other properties to set, they are set together in a single function call.

        If an argument is unspecified or None, then the setting is unchanged."""
    pixel_format: PixelFormat
    """The pixel format of captured frames"""
    frame_size: FrameSize
    """The size of captured frames"""
    contrast: int
    """The sensor contrast.  Positive values increase contrast, negative values lower it. The total range is device-specific but is often from -2 to +2 inclusive."""
    brightness: int
    """The sensor brightness.  Positive values increase brightness, negative values lower it. The total range is device-specific but is often from -2 to +2 inclusive."""
    saturation: int
    """The sensor saturation.  Positive values increase saturation (more vibrant colors), negative values lower it (more muted colors).  The total range is device-specific but the value is often from -2 to +2 inclusive."""
    sharpness: int
    """The sensor sharpness.  Positive values increase sharpness (more defined edges), negative values lower it (softer edges).  The total range is device-specific but the value is often from -2 to +2 inclusive."""
    denoise: int
    """The sensor 'denoise' setting.  Any camera sensor has inherent 'noise', especially in low brightness environments. Software algorithms can decrease noise at the expense of fine detail.  A larger value increases the amount of software noise removal.  The total range is device-specific but the value is often from 0 to 10."""
    gain_ceiling: GainCeiling
    """The sensor 'gain ceiling' setting. "Gain" is an analog multiplier applied to the raw sensor data.  The 'ceiling' is the maximum gain value that the sensor will use. A higher gain means that the sensor has a greater response to light, but also makes sensor noise more visible."""
    quality: int
    """The 'quality' setting when capturing JPEG images.  This is similar to the quality setting when exporting a jpeg image from photo editing software.  Typical values range from 5 to 40, with higher numbers leading to larger image sizes and better overall image quality. However, when the quality is set to a high number, the total size of the JPEG data can exceed the size of an internal buffer, causing image capture to fail."""
    colorbar: bool
    """When `True`, a test pattern image is captured and the real sensor data is not used."""
    whitebal: bool
    """When `True`, the camera attempts to automatically control white balance.  When `False`, the `wb_mode` setting is used instead."""
    gain_ctrl: bool
    """When `True`, the camera attempts to automatically control the sensor gain, up to the value in the `gain_ceiling` property.  When `False`, the `agc_gain` setting is used instead."""
    exposure_ctrl: bool
    """When `True` the camera attempts to automatically control the exposure. When `False`, the `aec_value` setting is used instead."""
    hmirror: bool
    """When `True` the camera image is mirrored left-to-right"""
    vflip: bool
    """When `True` the camera image is flipped top-to-bottom"""
    aec2: bool
    """When `True` the sensor's "night mode" is enabled, extending the range of automatic gain control."""
    awb_gain: bool
    """Access the awb_gain property of the camera sensor"""
    agc_gain: int
    """Access the gain level of the sensor. Higher values produce brighter images.  Typical settings range from 0 to 30. """
    aec_value: int
    """Access the exposure value of the camera.  Higher values produce brighter images.  Typical settings range from 0 to 1200."""
    special_effect: int
    """Enable a "special effect". Zero is no special effect.  On OV5640, special effects range from 0 to 6 inclusive and select various color modes."""
    wb_mode: int
    """The white balance mode.  0 is automatic white balance.  Typical values range from 0 to 4 inclusive."""
    ae_level: int
    """The exposure offset for automatic exposure.  Typical values range from -2 to +2."""
    dcw: bool
    """When `True` an advanced white balance mode is selected."""
    bpc: bool
    """When `True`, "black point compensation" is enabled. This can make black parts of the image darker."""
    wpc: bool
    """When `True`, "white point compensation" is enabled.  This can make white parts of the image whiter."""
    raw_gma: bool
    """When `True`, raw gamma mode is enabled."""
    lenc: bool
    """Enable "lens correction". This can help compensate for light fall-off at the edge of the sensor area."""
    max_frame_size: FrameSize
    """The maximum frame size that can be captured"""
    address: int
    """The I2C (SCCB) address of the sensor"""
    sensor_name: str
    """The name of the sensor"""
    supports_jpeg: bool
    """True if the sensor can capture images in JPEG format"""
    height: int
    """The height of the image being captured"""
    width: int
    """The width of the image being captured"""
    grab_mode: GrabMode
    """The grab mode of the camera"""
    framebuffer_count: int
    """True if double buffering is used"""
